@@ -45,8 +45,6 @@ fn example_xdp_probe(ctx: XdpContext) -> XdpResult {
 extern crate proc_macro;
 extern crate proc_macro2;
 use proc_macro::TokenStream;
-#[cfg(RUSTC_IS_NIGHTLY)]
-use proc_macro::{Diagnostic, Level};
 use proc_macro2::{Ident, Span, TokenStream as TokenStream2};
 use quote::quote;
 use std::str;
@@ -266,6 +264,8 @@ pub fn map(attrs: TokenStream, item: TokenStream) -> TokenStream {
         let mod_ident = syn::Ident::new(&mod_name, static_item.ident.span());
         let ktype = key_type.unwrap();
         let vtype = value_type.unwrap();
+        let map_btf_name = format!("MAP_BTF_{}", static_item.ident.to_string());
+        let map_btf_ident = syn::Ident::new(&map_btf_name, static_item.ident.span());
         tokens.extend(quote! {
             mod #mod_ident {
                 #[allow(unused_imports)]
@@ -280,9 +280,9 @@ pub fn map(attrs: TokenStream, item: TokenStream) -> TokenStream {
                 // `impl Sync` is needed to allow pointer types of keys and values
                 unsafe impl Sync for MapBtf {}
                 const N: usize = mem::size_of::<MapBtf>();
-                #[used]
-                #[link_section = #section_name]
-                static MAP_BTF: MapBtf = unsafe { mem::transmute::<[u8; N], MapBtf>([0u8; N])};
+                #[no_mangle]
+                #[link_section = "maps.ext"]
+                static #map_btf_ident: MapBtf = unsafe { mem::transmute::<[u8; N], MapBtf>([0u8; N]) };
             }
         });
     }
@@ -309,13 +309,65 @@ fn probe_impl(ty: &str, attrs: TokenStream, item: ItemFn, mut name: String) -> T
     tokens.into()
 }
 
+fn probe_pair_impl(pre: &str, attrs: TokenStream, item: ItemFn, mut name: String) -> TokenStream {
+    if !attrs.is_empty() {
+        name = match parse_macro_input!(attrs as Expr) {
+            Expr::Lit(ExprLit {
+                lit: Lit::Str(s), ..
+            }) => s.value(),
+            _ => panic!("expected string literal"),
+        }
+    };
+
+    let ident = item.sig.ident.clone();
+    let map_ident = Ident::new(&format!("PARMS_{}", ident), Span::call_site());
+    let enter_ident = Ident::new(&format!("enter_{}", ident), Span::call_site());
+    let exit_ident = Ident::new(&format!("exit_{}", ident), Span::call_site());
+    let probe_ident = Ident::new(&format!("{}probe", pre), Span::call_site());
+    let retprobe_ident = Ident::new(&format!("{}retprobe", pre), Span::call_site());
+
+    let tokens = quote! {
+        #[map]
+        static mut #map_ident: HashMap<u64, [u64; 5]> = HashMap::with_max_entries(10240);
+
+        #[#probe_ident(#name)]
+        fn #enter_ident(regs: Registers) {
+            let pid_tgid = bpf_get_current_pid_tgid();
+            let parms = [regs.parm1(), regs.parm2(), regs.parm3(), regs.parm4(), regs.parm5()];
+            unsafe {
+                #map_ident.set(&pid_tgid, &parms);
+            }
+        }
+
+        #[#retprobe_ident(#name)]
+        fn #exit_ident(regs: Registers) {
+            let pid_tgid = bpf_get_current_pid_tgid();
+            let parms = unsafe {
+                match #map_ident.get(&pid_tgid) {
+                    Some(parms) => {
+                        let parms = *parms;
+                        #map_ident.delete(&pid_tgid);
+                        parms
+                    }
+                    None => return,
+                }
+            };
+            let _ = unsafe { #ident(regs, parms) };
+        }
+
+        #item
+    };
+
+    tokens.into()
+}
+
 fn wrap_kprobe(item: ItemFn) -> ItemFn {
     let ident = item.sig.ident.clone();
     let outer_ident = Ident::new(&format!("outer_{}", ident), Span::call_site());
     parse_quote! {
         fn #outer_ident(ctx: *mut c_void) -> i32 {
             let regs = ::redbpf_probes::registers::Registers::from(ctx);
-            let _ = #ident(regs);
+            let _ = unsafe { #ident(regs) };
             return 0;
 
             #item
@@ -345,6 +397,7 @@ pub fn kprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
 /// Attribute macro that must be used to define [`kretprobes`](https://www.kernel.org/doc/Documentation/kprobes.txt).
 ///
 /// # Example
+///
 /// ```no_run
 /// use redbpf_probes::kprobe::prelude::*;
 ///
@@ -353,10 +406,44 @@ pub fn kprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
 ///     // this is executed when clone() returns
 /// }
 /// ```
+///
+/// # Function parameters
+///
+/// In general, the `parmX` methods of the `regs` argument do **not**
+/// return the original parameter values that the function being probed
+/// was called with. The reason is that those parameters are passed
+/// via (architecture-dependent) general purpose registers, and the
+/// function code most likely overwrites some or all of those registers.
+/// RedBPF provides a convenient way to access original function parameters
+/// by declaring the retprobe with an additional array argument that
+/// receives function parameters 1-5:
+///
+/// ```no_run
+/// use redbpf_probes::kprobe::prelude::*;
+///
+/// #[kretprobe("__x64_sys_clone")]
+/// fn clone_exit(regs: Registers, parms: [u64; 5]) {
+///     // this is executed when clone() returns
+/// }
+/// ```
+///
+/// To make this possible, RedBPF generates a global map, and an entry probe
+/// corresponding to the retprobe which stores the original parameters
+/// in that map. A generated retprobe wrapper retrieves the parameters from
+/// the map, and calls the provided function with the parameter array as
+/// an argument.
+///
+/// Note that if no parameters for the current thread are found in the map
+/// (for example because the capacity of the map has been exhausted, or
+/// the retprobe was registered after the function had already been entered),
+/// **the retprobe is not called** for that function invocation.
 #[proc_macro_attribute]
 pub fn kretprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
     let item = parse_macro_input!(item as ItemFn);
     let name = item.sig.ident.to_string();
+    if item.sig.inputs.len() == 2 {
+        return probe_pair_impl("k", attrs, item, name);
+    }
     let wrapper = wrap_kprobe(item);
     probe_impl("kretprobe", attrs, wrapper, name)
 }
@@ -383,6 +470,7 @@ pub fn uprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
 /// Attribute macro that must be used to define [`uretprobes`](https://www.kernel.org/doc/Documentation/trace/uprobetracer.txt).
 ///
 /// # Example
+///
 /// ```no_run
 /// use redbpf_probes::uprobe::prelude::*;
 ///
@@ -391,10 +479,44 @@ pub fn uprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
 ///     // this is executed when getaddrinfo() returns
 /// }
 /// ```
+///
+/// # Function parameters
+///
+/// In general, the `parmX` methods of the `regs` argument do **not**
+/// return the original parameter values that the function being probed
+/// was called with. The reason is that those parameters are passed
+/// via (architecture-dependent) general purpose registers, and the
+/// function code most likely overwrites some or all of those registers.
+/// RedBPF provides a convenient way to access original function parameters
+/// by declaring the retprobe with an additional array argument that
+/// receives function parameters 1-5:
+///
+/// ```no_run
+/// use redbpf_probes::uprobe::prelude::*;
+///
+/// #[uretprobe]
+/// fn getaddrinfo(regs: Registers, parms: [u64; 5]) {
+///     // this is executed when getaddrinfo() returns
+/// }
+/// ```
+///
+/// To make this possible, RedBPF generates a global map, and an entry probe
+/// corresponding to the retprobe which stores the original parameters
+/// in that map. A generated retprobe wrapper retrieves the parameters from
+/// the map, and calls the provided function with the parameter array as
+/// an argument.
+///
+/// Note that if no parameters for the current thread are found in the map
+/// (for example because the capacity of the map has been exhausted, or
+/// the retprobe was registered after the function had already been entered),
+/// **the retprobe is not called** for that function invocation.
 #[proc_macro_attribute]
 pub fn uretprobe(attrs: TokenStream, item: TokenStream) -> TokenStream {
     let item = parse_macro_input!(item as ItemFn);
     let name = item.sig.ident.to_string();
+    if item.sig.inputs.len() == 2 {
+        return probe_pair_impl("u", attrs, item, name);
+    }
     let wrapper = wrap_kprobe(item);
     probe_impl("uretprobe", attrs, wrapper, name)
 }
@@ -424,7 +546,7 @@ pub fn xdp(attrs: TokenStream, item: TokenStream) -> TokenStream {
     let wrapper = parse_quote! {
         fn #outer_ident(ctx: *mut ::redbpf_probes::bindings::xdp_md) -> ::redbpf_probes::xdp::XdpAction {
             let ctx = ::redbpf_probes::xdp::XdpContext { ctx };
-            return match #ident(ctx) {
+            return match unsafe { #ident(ctx) } {
                 Ok(action) => action,
                 Err(_) => ::redbpf_probes::xdp::XdpAction::Pass
             };
@@ -460,7 +582,7 @@ pub fn socket_filter(attrs: TokenStream, item: TokenStream) -> TokenStream {
     let wrapper = parse_quote! {
         fn #outer_ident(skb: *const ::redbpf_probes::bindings::__sk_buff) -> i32 {
             let skb = ::redbpf_probes::socket_filter::SkBuff { skb };
-            return match #ident(skb) {
+            return match unsafe { #ident(skb) } {
                 Ok(::redbpf_probes::socket_filter::SkBuffAction::SendToUserspace) => -1,
                 _ => 0
             };
@@ -503,7 +625,7 @@ pub fn stream_parser(attrs: TokenStream, item: TokenStream) -> TokenStream {
         fn #outer_ident(skb: *const ::redbpf_probes::bindings::__sk_buff) -> i32 {
             let skb = ::redbpf_probes::socket::SkBuff { skb };
             use ::redbpf_probes::sockmap::StreamParserAction::*;
-            return match #ident(skb) {
+            return match unsafe { #ident(skb) } {
                 Ok(MessageLength(len)) if len > 0 => len as i32,
                 Ok(MoreDataWanted) => 0,
                 Ok(SendToUserspace) => -86,  // -ESTRPIPE
@@ -546,7 +668,7 @@ pub fn stream_verdict(attrs: TokenStream, item: TokenStream) -> TokenStream {
             let skb = ::redbpf_probes::socket::SkBuff { skb };
             use ::redbpf_probes::socket::SkAction;
 
-            return match #ident(skb) {
+            return match unsafe { #ident(skb) } {
                 SkAction::Pass => ::redbpf_probes::bindings::sk_action_SK_PASS,
                 SkAction::Drop => ::redbpf_probes::bindings::sk_action_SK_DROP,
             } as i32;
@@ -568,7 +690,7 @@ pub fn tc_action(attrs: TokenStream, item: TokenStream) -> TokenStream {
     let wrapper = parse_quote! {
         fn #outer_ident(skb: *const ::redbpf_probes::bindings::__sk_buff) -> i32 {
             let skb = ::redbpf_probes::socket::SkBuff { skb };
-            return match #ident(skb) {
+            return match unsafe { #ident(skb) } {
                 Ok(act) => act as i32,
                 Err(_) => -1
             };
@@ -579,3 +701,25 @@ pub fn tc_action(attrs: TokenStream, item: TokenStream) -> TokenStream {
 
     probe_impl("tc_action", attrs, wrapper, name)
 }
+
+/// Attribute macro for defining a BPF iterator of `task`
+#[proc_macro_attribute]
+pub fn task_iter(attrs: TokenStream, item: TokenStream) -> TokenStream {
+    let item = parse_macro_input!(item as ItemFn);
+    let name = item.sig.ident.to_string();
+    let ident = item.sig.ident.clone();
+    let outer_ident = Ident::new(&format!("outer_{}", ident), Span::call_site());
+    let wrapper = parse_quote! {
+        fn #outer_ident(ctx: *mut bpf_iter__task) -> i32 {
+            let task_ctx = ::redbpf_probes::bpf_iter::context::TaskIterContext { ctx };
+            return match unsafe { #ident(task_ctx) } {
+                BPFIterAction::Ok => 0,
+                BPFIterAction::Retry => 1,
+            };
+
+            #item
+        }
+    };
+
+    probe_impl("task_iter", attrs, wrapper, name)
+}
